"""MPRIS D-Bus for a remote Mopidy instance.

Control a remote Mopidy instance through DBus, e.g. with sound applets.

Many desktop environments can control a media player through the
D-Bus (http://www.freedesktop.org/wiki/Software/dbus) via the
Media Player Remote Interfacing Specification (MPRIS)
(https://specifications.freedesktop.org/mpris-spec/latest/) interface. To
control a locally running instance of Mopidy, there exists the `Mopidy-MPRIS`
package:
- https://mopidy.com/ext/mpris/
- https://pypi.org/project/Mopidy-MPRIS
- https://github.com/mopidy/mopidy-mpris

For this to work, Mopidy has to run on the same computer as the the D-Bus. If
Mopidy runs on a remote server, there was only the following solution
available:
1. Install and enable the MPD interface for Mopidy:
    - https://mopidy.com/ext/mpd/
    - https://pypi.org/project/Mopidy-MPD/
    - https://github.com/mopidy/mopidy-mpd
2. Install [`mpDris2`](https://github.com/eonpatapon/mpDris2).
3. Start `mpDris2`, whenever you log into your desktop environment.

`mopidris` connects directly to the remote Mopidy without the detour through
the `mpd` interface.


Optional packages
=================

You can install the following optional packages. `mopidris` makes use of them,
if it can find them, but it also works well without them:
1. `aiorun`: If the event loop is terminated with CTRL-C, sometimes the port
   is not released and 'mopidris' cannot be restarted. 'aiorun' can help here
   (but it can still happen, that the DBUS hangs ...)
    - https://pypi.org/project/aiorun/
    - https://github.com/cjrh/aiorun
2. `colored_traceback`: Shows coloured tracebacks, to help debugging
   exceptions.
    - https://pypi.org/project/colored-traceback
    - https://github.com/staticshock/colored-traceback.py
3. `colorlog`: Colour the log to make it easier to read.
    - https://pypi.org/project/colorlog/
    - https://github.com/borntyping/python-colorlog


Usage
=====

'__main__.py' contains the logic for being called from a command line, so this
package can be called as `python3 mopidris`. For more information see
`python3 mopidris --help`, '__main__.py' and the README file.


Created on Sun Mar 28 18:18:06 2021


Copyright (C) 2021  Stephan Helma

This file is part of mopidris.

mopidris is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mopidris is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mopidris. If not, see <https://www.gnu.org/licenses/>.

"""

import logging

import dbus_next as dbus
from mopidy_asyncio_client import MopidyClient


logger = logging.getLogger(__name__)


async def main_loop(host=None, port=None,
                    interfaces=[],
                    can_quit=False,
                    mangle=True,
                    loop_type='asyncio',
                    debug=False,
                    **kwargs):
    """Asyncio's main loop function.

    Parameters
    ----------
    host : str, optional
        The host or ip address, where the Mopidy instance can be found.
        The default is 'None', using 'mopidy_asyncio_client.MopidyClient''s
        default host ("localhost").
    port : int, optional
        The port number, where the Mopidy instance can be found.
        The default is 'None', using 'mopidy_asyncio_client.MopidyClient''s
        default port number (6680).
    interfaces : [str]
        A list with the additional interfaces to be activated. Valid values
        "tracklist" and "playlists".
    loop_type : str, optional
        The type of the loop. Currently supported are "aiorun" and the
        standard "asyncio".
        The default is "asyncio".
    can_quit : bool
        If the program advertise, that it can be quit.
    mangle : bool
        If the titles of Internet radio streams should be split at " - " into
        artist and title.
    debug : bool, optional
        If debug output should be generated.
        The default is 'False'.self._decoder = None
    **kwargs
        Further keyword arguments passed on to 'MopidyClient()'.

    Returns
    -------
    None.

    """
    # We could have incorporated these keywords into '**kwargs', but we want
    # to have them stand out...
    if host is not None:
        kwargs['host'] = host
    if port is not None:
        kwargs['port'] = port

    bus_name = f"RemoteMopidy._{host.replace('.', '_')}__{port}"

    async with MopidyClient(**kwargs) as mopidy:

        logger.info('%r', mopidy)

        #
        # DBus message bus
        #

        bus = await dbus.aio.MessageBus().connect()
        logger.debug('%r', bus)

        #
        # DBus Interfaces
        #
        # Only import interfaces, which were specified

        # Import and initiate them
        from interfaces.root import MopidyInterface
        root_interface = MopidyInterface(
            mopidy, can_quit=can_quit, loop_type=loop_type)

        from interfaces.player import MopidyPlayerInterface
        player_interface = MopidyPlayerInterface(
            mopidy, mangle=mangle)

        if 'tracklist' in interfaces:
            from interfaces.tracklist import MopidyTrackListInterface
            tracklist_interface = MopidyTrackListInterface(
                mopidy, player_interface)

        if 'playlists' in interfaces:
            from interfaces.playlists import MopidyPlaylistsInterface
            playlists_interface = MopidyPlaylistsInterface(
                mopidy)

        # Make them available on the DBUS
        bus.export('/org/mpris/MediaPlayer2', root_interface)
        bus.export('/org/mpris/MediaPlayer2', player_interface)
        if 'tracklist' in interfaces:
            bus.export('/org/mpris/MediaPlayer2', tracklist_interface)
        if 'playlists' in interfaces:
            bus.export('/org/mpris/MediaPlayer2', playlists_interface)

        await bus.request_name(f'org.mpris.MediaPlayer2.{bus_name}')

        #
        # Bind Mopidy events to the the org.mpris.MediaPlayer2 interfaces.
        #
        # Available Mopidy events are described in:
        #   https://docs.mopidy.com/en/latest/api/core/#core-events
        #

        # Catch all
        if debug:
            # Mopidy's catch all 'on_event(event, **kwargs)' is exposed as
            # '*' in 'mopidy_asyncio_client'; only catch it, if we are in debug
            # mode - and do it first, so that we get the event displayed
            # before it is processed by the special event listeners
            mopidy.bind('*', root_interface._debug_event)

        # Bind events for org.mpris.MediaPlayer2 interface; yes, there are none
        for event in ():
            mopidy.bind(
                event,
                getattr(root_interface, f'on_mopidyevent_{event}'))

        # Bind events for org.mpris.MediaPlayer2.Player interface
        for event in ('mute_changed',
                      'options_changed',
                      'playback_state_changed', 'stream_title_changed',
                      'seeked',
                      'volume_changed'):
            mopidy.bind(
                event,
                getattr(player_interface, f'on_mopidyevent_{event}'))

        # Bind events for org.mpris.MediaPlayer2.TrackLists interface
        if 'tracklist' in interfaces:
            for event in ('track_playback_ended', 'track_playback_paused',
                          'track_playback_resumed', 'track_playback_started',
                          'tracklist_changed'):
                mopidy.bind(
                    event,
                    getattr(tracklist_interface, f'on_mopidyevent_{event}'))

        # Bind events for org.mpris.MediaPlayer2.Playlists interface
        if 'playlists' in interfaces:
            for event in ('playlist_changed', 'playlist_deleted',
                          'playlists_loaded'):
                mopidy.bind(
                    event,
                    getattr(playlists_interface, f'on_mopidyevent_{event}'))

        await bus.wait_for_disconnect()



"""The Playlists MPRIS D-Bus interface for a remote Mopidy instance.

The D-Bus name of this interface is 'org.mpris.MediaPlayer2.Playlists'.


Created on Sun Mar 28 18:18:06 2021


Copyright (C) 2021  Stephan Helma

This file is part of mopidris.

mopidris is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mopidris is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mopidris. If not, see <https://www.gnu.org/licenses/>.

"""

import logging

from dbus_next.service import method, dbus_property, PropertyAccess, signal

from .base import _MopidyBaseInterface


logger = logging.getLogger(__name__)


class MopidyPlaylistsInterface(_MopidyBaseInterface):
    """The "org.mpris.MediaPlayer2.Playlists" interface.

    Provides access to the media player's playlists.

    Since D-Bus does not provide an easy way to check for what interfaces are
    exported on an object, clients should attempt to get one of the properties
    on this interface to see if it is implemented.

    See:
        https://specifications.freedesktop.org/mpris-spec/2.2/Playlists_Interface.html
    """

    def __init__(self, mopidy, *args, **kwargs):
        """Initialize the 'MopidyPlaylistsInterface' class.

        Parameters
        ----------
        mopdiy : mopidy_asyncio_client.MopidyClient
            The mopidy client
        *args : arguments
            Passed on to parent class.
        **kwargs : keyword arguments
            Passed on to parent class.

        Returns
        -------
        None.

        """
        super().__init__('org.mpris.MediaPlayer2.Playlists',
                         mopidy, *args, **kwargs)

    #
    # Mopidy events
    #

    async def on_mopidyevent_playlist_changed(self, data):
        """Is called whenever a playlist is changed.

        Parameters
        ----------
        data['playlist'] : mopidy.models.Playlist
            The changed playlist.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.Playlists.playlist_changed'
            '(data=%s) received',
            data)
        self.PlaylistChanged(data['playlist'])

    async def on_mopidyevent_playlist_deleted(self, data):
        """Is called whenever a playlist is deleted.

        Parameters
        ----------
        data['uri'] : str
            The changed playlist.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.Playlists.mute_deleted'
            '(data=%s) received',
            data)
        count = await self._get_PlaylistCount()
        self.emit_properties_changed(
            PlaylistCount=count)

    async def on_mopidyevent_playlists_loaded(self, data):
        """Is called whenever a playlist is loaded or refreshed.

        Parameters
        ----------
        data : {}
            No data supplied, hence an empty dictionary.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.Playlists.mute_loaded'
            '(data=%s) received',
            data)
        count = await self._get_PlaylistCount()
        self.emit_properties_changed(
            PlaylistCount=count)

    #
    # Methods
    #

    # ActivatePlaylist(o: PlaylistId)                   → nothing

    @method()
    async def ActivatePlaylist(self, playlist_id: 'o'):
        """Start playing the given playlist.

        This must be implemented! If the media player does not allow clients
        to change the playlist, it should not implement this interface at all.

        It is up to the media player whether this completely replaces the
        current tracklist, or whether it is merely inserted into the tracklist
        and the first track starts. For example, if the media player is
        operating in a "jukebox" mode, it may just append the playlist to the
        list of upcoming tracks, and skip to the first track in the playlist.

        Parameters
        ----------
        playlist_id : PlaylistId
            The id of the playlist to activate.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Playlists.ActivatePlaylist'
            '(playlist_id=%s) called',
            playlist_id)
        playlist_uri = self.dbusid_to_playlistid(playlist_id)
        playlist = await self.mopidy.playlists.lookup(playlist_uri)
        if playlist and playlist.tracks:
            tl_tracks = await self.mopidy.tracklist.add(playlist.tracks)
            await self.player.play(tlid=tl_tracks[0].tlid)

    # GetPlaylists(u: Index, u: MaxCount, s: Order, b: ReverseOrder)
    #                                                   → a(oss): Playlists

    @method()
    async def GetPlaylists(self,
                           index: 'u', max_count: 'u',
                           order: 's', reverse_order: 'b') -> 'a(oss)':
        """Get a set of playlists.

        Parameters
        ----------
        index : int
            The index of the first playlist to be fetched (according to the
            ordering).
        max_count : int
            The maximum number of playlists to fetch.
        order : PlaylistOrdering
            The ordering that should be used.
        reverse_order : bool
            'False':
                The order is not reversed.
            'True':
                The order should be reversed.

        Returns
        -------
        [PlayList]
            A list of (at most 'max_count') playlists.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Playlists.GetPlaylists'
            '(index=%s, max_count=%s, order=%s, reverse_order=%s) called',
            index, max_count, order, reverse_order)
        # Get playlists
        playlists = await self.mopidy.playlists.as_list()
        # Order playlists
        if order == 'Alphabetical':
            # Order by playlist.name
            playlists.sort(key=lambda p: p.name, reverse=reverse_order)
        elif order == 'User':
            # Do not change order
            if reverse_order:
                playlists.reverse()
            else:
                pass
        # Return requested slice
        return [(self.playlist_to_dbusid(playlists), playlists.name, '')
                for playlist in playlists[index:index+max_count]]

    #
    # Signals
    #

    # PlaylistChanged((oss): Playlist)

    @signal()
    def PlaylistChanged(self, playlist: '(oss)') -> '(oss)':
        """Indicate that 'Name' or 'Icon' attribute of a playlist has changed.

        Indicates that either the 'Name' or 'Icon' attribute of a playlist has
        changed.

        Client implementations should be aware that this signal may not be
        implemented.

        Parameters
        ----------
        playlist : PlayList
            The playlist which details have changed.

        Returns
        -------
        PlayList
            The playlist which details have changed.

        """
        logger.info(
            'Signal org.mpris.MediaPlayer2.Playlists.PlaylistChanged'
            '(playlist=%s) to emit',
            playlist)
        return playlist

    #
    # Properties
    #

    # PlaylistCount         u                       Read only

    @dbus_property(access=PropertyAccess.READ)
    async def PlaylistCount(self) -> 'u':                       # noqa: D401
        """The number of playlists available.

        Signals
        ----------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        int
            The number of playlists available.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Playlists.PlaylistCount '
            'requested')
        return await self._get_PlaylistCount()

    async def _get_PlaylistCount(self):
        playlists = await self.mopidy.playlists.as_list()
        return len(playlists)

    # Orderings             as (Playlist_Ordering_List)     Read only

    @dbus_property(access=PropertyAccess.READ)
    def Orderings(self) -> 'as':                                # noqa: D401
        """The available orderings.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        [PlaylistOrdering]
            The available orderings. At least one must be offered.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Playlists.Orderings '
            'requested')
        return self._get_Orderings()

    def _get_Orderings(self):
        # Hard coded as in 'mopidy_mpris/playlists.py'
        return [
            "Alphabetical",     # Order by playlist.name
            "User"]             # Don't change order

    # ActivePlaylist        (b(oss)) (Maybe_Playlist)       Read only

    @dbus_property(access=PropertyAccess.READ)
    def ActivePlaylist(self) -> '(b(oss))':               # noqa: D401
        """The currently-active playlist.

        Note that this may not have a value even after 'ActivatePlaylist()'
        is called with a valid playlist id as 'ActivatePlaylist()'
        implementations have the option of simply inserting the contents of
        the playlist into the current tracklist.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        [MaybePlaylist]
            List of 'MaybePlaylist' items.

            If there is no currently-active playlist, the structure's 'valid'
            field will be 'False', and the 'Playlist' details are undefined.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Playlists.ActivePlaylist '
            'requested')
        return self._get_ActivePlaylist()

    def _get_ActivePlaylist(self):
        # Hard coded as in 'mopidy_mpris/playlists.py'
        return [False, ['/', 'None', '']]



"""The root MPRIS D-Bus interface for a remote Mopidy instance.

The D-Bus name of this interface is 'org.mpris.MediaPlayer2'.


Created on Sun Mar 28 18:18:06 2021


Copyright (C) 2021  Stephan Helma

This file is part of mopidris.

mopidris is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mopidris is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mopidris. If not, see <https://www.gnu.org/licenses/>.

"""

import logging
import asyncio

import dbus_next as dbus
from dbus_next.service import method, dbus_property, PropertyAccess

from .base import _MopidyBaseInterface


logger = logging.getLogger(__name__)


class MopidyInterface(_MopidyBaseInterface):
    """The "org.mpris.MediaPlayer2" interface.

    See:
        https://specifications.freedesktop.org/mpris-spec/2.2/Media_Player.html
    """

    def __init__(self, mopidy, *args,
                 can_quit=False, loop_type='asyncio', **kwargs):
        """Inititalize the 'MopidyInterface' class.

        Parameters
        ----------
        mopdiy : mopidy_asyncio_client.MopidyClient
            The mopidy client
        can_quit : bool, optional
            If the program advertise, that it can be quit.
            The default is 'False'.
        loop_type : str, optional
            The type of the loop. Currently supported are "aiorun" and the
            standard "asyncio".
            The default is "asyncio".
        *args : arguments
            Passed on to parent class.
        **kwargs : keyword arguments
            Passed on to parent class.

        Returns
        -------
        None.

        """
        self._can_quit = can_quit
        logger.info(
            f'The program {"can not" if can_quit else "CAN"} quit')
        self._loop_type = loop_type
        logger.debug(
            f'The asyncio library used is: {loop_type}')
        super().__init__('org.mpris.MediaPlayer2',
                         mopidy, *args, **kwargs)

    #
    # Mopidy events
    #

    #
    # Methods
    #

    # Raise                                             → nothing

    @method()
    def Raise(self):
        """Bring the media player's user interface to the front.

        The media player may be unable to control how its user interface is
        displayed, or it may not have a graphical user interface at all. In
        this case, the 'CanRaise' property is 'False' and this method does
        nothing.

        Raises
        ------
        NotSupported
            May be raised, if the 'CanRaise' property is 'False'.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Raise() called')
        if self._get_CanRaise():
            pass
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot raise a window, '
                f'because mopidy is running remotely')

    # Quit()                                            → nothing

    @method()
    async def Quit(self):
        """Cause the media player to stop running.

        The media player may refuse to allow clients to shut it down. In this
        case, the 'CanQuit' property is 'False' and this method does nothing.

        Raises
        ------
        NotSupported
            May be raised, if the 'CanQuit' property is 'False'.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Quit() called')
        if self._get_CanQuit():
            # Display quitting message
            # FIXME: Changed metadata is not displayed
            self.emit_properties_changed(
                Metadata={
                    'xesam:title': dbus.Variant('s', 'Quitting...'),
                    'xesam:artist': dbus.Variant('as', ['Quitting…'])})
            loop = asyncio.get_running_loop()
            if self._loop_type == 'aiorun':
                logging.debug(
                    "Shutting down using 'aiorun'")
                loop.stop()
            else:
                logging.debug(
                    "Shutting down using 'asyncio'")
                current_task = asyncio.current_task(loop)
                # Get running tasks
                tasks = [task
                         for task in asyncio.all_tasks(loop)
                         if task is not current_task]
                # Cancel all running tasks
                [task.cancel() for task in tasks]
                # Wait for the tasks to finish
                await asyncio.gather(*tasks)
                # Exit the loop
                loop.close()
            logging.info(
                'Shut down')
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot quit, '
                f'because mopidy is running remotely')

    #
    # Properties
    #

    # CanQuit               b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    def CanQuit(self) -> 'b':
        """Whether the media player can quit.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                Calling 'Quit()' will have no effect, and may raise a
                'NotSupported' error.
            'True':
                Calling 'Quit()' will cause the media application to attempt
                to quit (although it may still be prevented from quitting by
                the user, for example).

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.CanQuit requested')
        return self._get_CanQuit()

    def _get_CanQuit(self):
        return self._can_quit

    # Fullscreen            b                       Read/Write    (optional)

    @dbus_property(access=PropertyAccess.READWRITE)
    def Fullscreen(self) -> 'b':
        """Whether the media player is occupying the fullscreen.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Parameters
        ----------
        value : bool
            'False':
                Clients may set this property to 'False' to return to
                windowed mode.
            'True':
                Clients may set this property to 'True' to tell the media
                player to enter fullscreen mode.

            If 'CanSetFullscreen' is 'False', then attempting to set this
            property should have no effect, and may raise an error.
            However, even if it is 'True', the media player may still be
            unable to fulfil the request, in which case attempting to set
            this property will have no effect (but should not raise an
            error).

        Raises
        ------
        NotSupported
            May be raised, if the 'CanSetFullscreen' property is 'False'.

        Returns
        -------
        bool
            'False':
                The media player is not taking up the full screen.
            'True':
                The media player is taking up the full screen.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Fullscreen requested')
        return self._get_Fullscreen()

    @Fullscreen.setter
    def Fullscreen(self, value: 'b'):
        logger.info(
            'Property org.mpris.MediaPlayer2.Fullscreen set to: %s',
            value)
        try:
            self._set_Fullscreen(value)
        except NotImplementedError:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot occupy fullscreen, '
                'because mopidy is running remotely')
            return False

    def _get_Fullscreen(self):
        return False

    def _set_Fullscreen(self, value):
        if self._get_CanSetFullscreen():
            self.emit_properties_changed(
                Fullscreen=bool(value))
        else:
            raise NotImplementedError

    # CanSetFullscreen      b                       Read only     (optional)

    @dbus_property(access=PropertyAccess.READ)
    def CanSetFullscreen(self) -> 'b':
        """Whether the media player can occupy the fullscreen.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                Attempting to set 'Fullscreen' will have no effect, and may
                raise an error.
            'True':
                Attempting to set 'Fullscreen' will not raise an error, and
                (if it is different from the current value) will cause the
                media player to attempt to enter or exit fullscreen mode.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.CanSetFullscreen requested')
        return self._get_CanSetFullScreen()

    def _get_CanSetFullScreen(self):
        return False

    # CanRaise              b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    def CanRaise(self) -> 'b':
        """Whether the media application can be raised.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                Calling 'Raise()' will have no effect, and may raise a
                'NotSupported' error.
            'True':
                Calling 'Raise()' will cause the media application to attempt
                to bring its user interface to the front, although it may be
                prevented from doing so (by the window manager, for example).

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.CanRaise requested')
        return self._get_CanRaise()

    def _get_CanRaise(self):
        return False

    # HasTrackList          b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    def HasTrackList(self) -> 'b':
        """Whether the /org/mpris/MediaPlayer2 object implements
        the org.mpris.MediaPlayer2.TrackList interface.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                The /org/mpris/MediaPlayer2 object does not implement the
                org.mpris.MediaPlayer2.TrackList interface.
            'True':
                The /org/mpris/MediaPlayer2 object implements the
                org.mpris.MediaPlayer2.TrackList interface.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.HasTrackList requested')
        return self._get_HasTrackList()

    def _get_HasTrackList(self):
        return True

    # Identity              s                       Read only

    @dbus_property(access=PropertyAccess.READ)
    def Identity(self) -> 's':                                  # noqa: D401
        """A friendly name to identify the media player to users.

        This should usually match the name found in .desktop files.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        str
            The user friendly name of the media player.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Identity requested')
        return self._get_Identity()

    def _get_Identity(self):
        return str(self.mopidy)

    # DesktopEntry          s                       Read only     (optional)

    @dbus_property(access=PropertyAccess.READ)
    def DesktopEntry(self) -> 's':                              # noqa: D401
        """The basename of the installed .desktop file.

        Example
        -------
            The desktop entry file is '/usr/share/applications/vlc.desktop',
            and this property contains "vlc".

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        str
            The basename of an installed .desktop file which complies with
            the Desktop entry specification, with the ".desktop" extension
            stripped.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.DesktopEntry requested')
        return self._get_DesktopEntry()

    def _get_DesktopEntry(self):
        return 'mopidris'

    # SupportedUriSchemes   as                      Read only

    @dbus_property(access=PropertyAccess.READ)
    async def SupportedUriSchemes(self) -> 'as':                # noqa: D401
        """The URI schemes supported by the media player.

        This can be viewed as protocols supported by the player in almost all
        cases. Almost every media player will include support for the "file"
        scheme. Other common schemes are "http" and "rtsp".

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        [str]
            A list of the supported URI schemes (in lower-case).

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.SupportedUriSchemes requested')
        return await self._get_SupportedUriSchemes()

    async def _get_SupportedUriSchemes(self):
        return await self.mopidy.core.get_uri_schemes()

    # SupportedMimeTypes   as                       Read only

    @dbus_property(access=PropertyAccess.READ)
    def SupportedMimeTypes(self) -> 'as':                       # noqa: D401
        """The mime-types supported by the media player.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        [str]
            A list of supported mime-types in the standard format (e.g.
            "audio/mpeg" or "application/ogg")

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.SupportedMimeTypes requested')
        return self._get_SupportedMimeTypes()

    def _get_SupportedMimeTypes(self):
        # BUG: Supported MimeTypes are not exposed by mopidy
        # Issue:
        #   https://github.com/mopidy/mopidy/issues/812
        # PR has hard coded mime types:
        #   https://github.com/mopidy/mopidy-mpris/pull/11
        return [
            'audio/mpeg',
            'audio/x-ms-wma',
            'audio/x-ms-asf',
            'audio/x-flac',
            'audio/flac',
            'audio/l16;channels=2;rate=44100',
            'audio/l16;rate=44100;channels=2']



"""A base MPRIS D-Bus interface for a remote Mopidy instance.

Control a remote Mopidy instance through D-Bus, e.g. with sound applets.

MPRIS D-Bus Interface Specification
===================================

The Media Player Remote Interfacing Specification is a standard D-Bus
interface which aims to provide a common programmatic API for controlling
media players.

It provides a mechanism for discovery, querying and basic playback control of
compliant media players, as well as a tracklist interface which is used to add
context to the active media item.

https://specifications.freedesktop.org/mpris-spec/latest/


Bus Name Policy
---------------

Each media player must request a unique bus name which begins with
"org.mpris.MediaPlayer2.". For example:
    - "org.mpris.MediaPlayer2.audacious"
    - "org.mpris.MediaPlayer2.vlc"
    - "org.mpris.MediaPlayer2.bmp"
    - "org.mpris.MediaPlayer2.xmms2"

This allows clients to list available media players (either already running or
which can be started via D-Bus activation)

In the case where the media player allows multiple instances running
simultaneously, each additional instance should request a unique bus name,
adding a dot and a unique identifier to its usual bus name, such as one based
on a UNIX process id. For example, this could be:
    - "org.mpris.MediaPlayer2.vlc.instance7389"

Note: According to the D-Bus specification, the unique identifier "must only
contain the ASCII characters '[A-Z][a-z][0-9]_-'" and "must not begin with a
digit".


Types
-----

Track_Id (object path) - Unique track identifier:

    If the media player implements the 'TrackList' interface and allows the
    same track to appear multiple times in the tracklist, this must be unique
    within the scope of the tracklist.

    Note that this should be a valid D-Bus object id, although clients should
    not assume that any object is actually exported with any interfaces at
    that path.

    Media players may not use any paths starting with "/org/mpris" unless
    explicitly allowed by this specification. Such paths are intended to have
    special meaning, such as "/org/mpris/MediaPlayer2/TrackList/NoTrack" to
    indicate "no track".

MetaData ({str: variant}):

    A mapping from metadata attribute names to values.

    Items:
        key (str):
            The name of the attribute; see the metadata page for guidelines
            on names to use.
        value (variant):
            The value of the attribute, in the most appropriate format.

    Notable keys:
        "mpris:trackid":
            Must always be present!

            Must be of D-Bus type "o". This contains a D-Bus path that
            uniquely identifies the track within the scope of the playlist.
            There may or may not be an actual D-Bus object at that path; this
            specification says nothing about what interfaces such an object
            may implement.
        "mpris:length":
            If the length of the track is known, it should be provided in this
            key. The length must be given in microseconds, and be represented
            as a signed 64-bit integer.
        "mpris:artUrl":
            If there is an image associated with the track, a URL for it may
            be provided using this key.
        "xesam:...":
            For other metadata, fields defined by the Xesam ontology
            (http://xesam.org/main/XesamOntology) should be used, prefixed by
            "xesam:". See the metadata page on the freedesktop.org wiki
            (http://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata)
            for a list of common fields.

PlaylistId (str):

    Unique playlist identifier.

    This is a D-Bus object id as that is the definitive way to have unique
    identifiers on D-Bus. It also allows for future optional expansions to the
    specification where tracks are exported to D-Bus with an interface similar
    to "org.gnome.UPnP.MediaItem2".

PlayList ([[id, name, icon]]):

    A 'list' describing a playlist.

    Items:
        id (PlaylistId):
            A unique identifier for the playlist.

            This should remain the same if the playlist is renamed.
        name (str):
            The name of the playlist, typically given by the user.
        icon (str):
            The URI of an (optional) icon.

MaybePlaylist ([valid, playlist]):

    A data structure describing a playlist, or nothing.

    Items:
        valid (bool):
            Whether this structure refers to a valid playlist.
        playlist (PlayList):
            The playlist, providing 'valid' is 'True', otherwise undefined.

            When constructing this type, it should be noted that the playlist
            ID must be a valid object path, or D-Bus implementations may
            reject it. This is true even when 'valid' is 'False'. It is
            suggested that "/" is used as the playlist ID in this case.

PlaylistOrdering (str):

    Specifies the ordering of playlist.

    Values:
        "Alphabetical" (Alphabetical):
            Alphabetical ordering by name, ascending.
        "CreationDate" (Created):
            Ordering by creation date, oldest first.
        "ModifiedDate" (Modified):
            Ordering by last modified date, oldest first.
        "LastPlayDate" (Played):
            Ordering by date of last playback, oldest first.
        "UserDefined" (User):
            A user-defined ordering.


Created on Sun Mar 28 18:18:06 2021


Copyright (C) 2021  Stephan Helma

This file is part of mopidris.

mopidris is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mopidris is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mopidris. If not, see <https://www.gnu.org/licenses/>.
"""

import logging

import dbus_next as dbus


logger = logging.getLogger(__name__)


class _MopidyBaseInterface(dbus.service.ServiceInterface):
    """Base class for all "org.mpris.MediaPlayer2" interfaces."""

    _dbusstub_track = '/com/mopidy/track/'
    _dbusstub_playlist = '/com/mopidy/playlist/'

    def __init__(self, interface_name, mopidy):
        """Initialize the '_MopidyInterface' class.

        Parameters
        ----------
        interface_name : str
            The name of the this interface as it appears to clients. Must be
            a valid interface name.
        mopdiy : mopidy_asyncio_client.MopidyClient
            The mopidy client

        Returns
        -------
        None.

        """
        super().__init__(interface_name)
        self.mopidy = mopidy

    async def _debug_event(self, *args, **kwargs):
        """Catch all events from 'mopidy_asyncio_client.MopidyClient'."""
        logger.debug('************ Event ************\n%s', args)

    def emit_properties_changed(self, **changed_properties):
        """Overload the 'emit_properties_changed()' method.

        The 'dbus.service.ServiceInterface.emit_properties_changed()' is
        overloaded
        1. so that this method logs automatically.
        2. and that keyword arguments are used instead of a dictionary


        Parameters
        ----------
        **changed_properties : keyword arguments
            Passed on to overwritten 'emit_properties_changed()'.

        Returns
        -------
        None.

        """
        logger.info(
            'Signal org.freedesktop.DBus.Properties.PropertiesChanged'
            '(%s) to emit',
            changed_properties)
        super().emit_properties_changed(changed_properties)

    def mopidyid_to_dbusid(self, tl_id):
        """Generate the full D-Bus track id from Mopidy's track id.

        Parameters
        ----------
        tl_id : int
            The Mopidy track id.

        Returns
        -------
        str
            The full D-Bus track id.

        """
        return self._dbusstub_track + str(tl_id)

    def tltrack_to_dbusid(self, tl_track):
        """Generate the full D-Bus track id from Mopidy's 'tl_track'.

        Parameters
        ----------
        tl_track : mopidy.TlTrack
            A dictionary with track information.

        Returns
        -------
        str
            The full D-Bus track id.

        """
        return self._dbusstub_track + str(tl_track["tlid"])

    def dbusid_to_mopidyid(self, tlid):
        """Convert a D-Bus track id to Mopidy's tl_id.

        Parameters
        ----------
        tlid : str
            A string starting with the D-Bus path '/com/mopidy/track/'.

        Raises
        ------
        ValueError:
            If the 'tlid' does not start with '/com/mopidy/track' or the id
            cannot be converted to in integer value.

        Returns
        -------
        int
            The Mopidy tracklist id.

        """
        if tlid.startswith(self._dbusstub_track):
            try:
                return int(tlid.replace(self._dbusstub_track))
            except ValueError as e:
                raise ValueError(
                    f'Could not convert the id to an integer. {e}')
        else:
            raise ValueError(
                f"The id does not start with '{self._dbusstub_track}'. "
                f'Got: {tlid}')

    def playlist_to_dbusid(self, playlist):
        """Generate the full D-Bus playlist id from Mopidy's 'playlist'.

        Parameters
        ----------
        playlist : mopidy.models.Ref
            The playlist.

        Returns
        -------
        str
            The full D-Bus playlist id.

        """
        return self._dbusstub_playlist + str(playlist.uri)

    def dbusid_to_playlistid(self, playlist_id):
        """Convert a D-Bus playlist id to Mopidy's playlist.

        Parameters
        ----------
        playlist_id : str
            A string starting with the D-Bus path '/com/mopidy/playlist/'.

        Raises
        ------
        ValueError
            If the 'playlist_id' does not start with '/com/mopidy/playlist/'.

        Returns
        -------
        str
            The Mopidy playlist id.

        """
        if playlist_id.startswith(self._dbusstub_playlist):
            return playlist_id.replace(self._dbusstub_playlist)
        else:
            raise ValueError(
                f"The id does not start with '{self._dbusstub_playlist}'. "
                f'Got: {playlist_id}')



"""The TrackList MPRIS D-Bus interface for a remote Mopidy instance.

The D-Bus name of this interface is 'org.mpris.MediaPlayer2.TrackList'.


Created on Sun Mar 28 18:18:06 2021


Copyright (C) 2021  Stephan Helma

This file is part of mopidris.

mopidris is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mopidris is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mopidris. If not, see <https://www.gnu.org/licenses/>.

"""

import logging

import dbus_next as dbus
from dbus_next.service import method, dbus_property, PropertyAccess, signal

from .base import _MopidyBaseInterface


logger = logging.getLogger(__name__)


class MopidyTrackListInterface(_MopidyBaseInterface):
    """The "org.mpris.MediaPlayer2.TrackList" interface.

    Provides access to a short list of tracks which were recently played or
    will be played shortly. This is intended to provide context to the
    currently-playing track, rather than giving complete access to the media
    player's playlist.

    Example use cases are the list of tracks from the same album as the
    currently playing song or the Rhythmbox play queue.

    Each track in the tracklist has a unique identifier. The intention is that
    this uniquely identifies the track within the scope of the tracklist. In
    particular, if a media item (a particular music file, say) occurs twice in
    the track list, each occurrence should have a different identifier. If a
    track is removed from the middle of the playlist, it should not affect the
    track ids of any other tracks in the tracklist.

    As a result, the traditional track identifiers of URLs and position in the
    playlist cannot be used. Any scheme which satisfies the uniqueness
    requirements is valid, as clients should not make any assumptions about
    the value of the track id beyond the fact that it is a unique identifier.

    Note that the (memory and processing) burden of implementing the
    'TrackList' interface and maintaining unique track ids for the playlist
    can be mitigated by only exposing a subset of the playlist when it is very
    long (the 20 or so tracks around the currently playing track, for example).
    This is a recommended practice as the tracklist interface is not designed
    to enable browsing through a large list of tracks, but rather to provide
    clients with context about the currently playing track.

    See:
        https://specifications.freedesktop.org/mpris-spec/2.2/Track_List_Interface.html#Property:CanEditTracks
    """

    def __init__(self, mopidy, player_interface, *args, **kwargs):
        """Initialize the 'MopidyTrackListInterface' class.

        Parameters
        ----------
        mopdiy : mopidy_asyncio_client.MopidyClient
            The mopidy client
        player_interface : MopidyPlayerInterface
            The D-Bus "org.mpris.MediaPlayer2.Player" interface.
        *args : arguments
            Passed on to parent class.
        **kwargs : keyword arguments
            Passed on to parent class.

        Returns
        -------
        None.

        """
        self.player_interface = player_interface
        super().__init__('org.mpris.MediaPlayer2.TrackList',
                         mopidy, *args, **kwargs)

    #
    # Mopidy events
    #

    async def on_mopidyevent_track_playback_ended(self, data):
        """Is called whenever playback of a track ends.

        Parameters
        ----------
        data['tl_track'] : mopidy.models.TlTrack
            The track that was played before playback stopped.
        data['time_position'] : int
            The time position in milliseconds.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event '
            'org.mpris.MediaPlayer2.TrackLists.track_playback_ended'
            '(data=%s) received',
            data)
        status = await self.player_interface._get_PlaybackStatus()
        metadata = await self.player_interface._get_Metadata(
            tl_track=data['tl_track'])
        self.player_interface.emit_properties_changed(
            PlaybackStatus=status,
            Metadata=metadata)

    async def on_mopidyevent_track_playback_paused(self, data):
        """Is called whenever playback of a track is paused.

        Parameters
        ----------
        data['tl_track'] : mopidy.models.TlTrack
            The track that was playing when playback paused.
        data['time_position'] : int
            The time position in milliseconds.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event '
            'org.mpris.MediaPlayer2.TrackLists.track_playback_paused'
            '(data=%s) received',
            data)
        self.player_interface.emit_properties_changed(
            PlaybackStatus='Paused')

    async def on_mopidyevent_track_playback_resumed(self, data):
        """Is called whenever playback of a track is resumed.

        Parameters
        ----------
        data['tl_track'] : mopidy.models.TlTrack
            The track that was playing when playback resumed.
        data['time_position'] : int
            The time position in milliseconds.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event '
            'org.mpris.MediaPlayer2.TrackLists.track_playback_resumed'
            '(data=%s) received',
            data)
        self.player_interface.emit_properties_changed(
            PlaybackStatus='Playing')

    async def on_mopidyevent_track_playback_started(self, data):
        """Is called whenever a new track starts playing.

        Parameters
        ----------
        data['tl_track'] : mopidy.models.TlTrack
            The track that just started playing.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event '
            'org.mpris.MediaPlayer2.TrackLists.track_playback_started'
            '(data=%s) received',
            data)
        metadata = await self.player_interface._get_Metadata(
            tl_track=data['tl_track'])
        self.player_interface.emit_properties_changed(
            PlaybackStatus='Playing',
            Metadata=metadata)

    async def on_mopidyevent_tracklist_changed(self, data):
        """Is called whenever the tracklist is changed.

        Parameters
        ----------
        data : {}
            No data supplied, hence an empty dictionary.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.TrackLists.tracklist_changed'
            '(data=%s) received',
            data)
        # Nothing to do here
        pass

    #
    # Methods
    #

    # GetTracksMetadata(ao: TrackIds)                   → aa{sv}: Metadata
    @method()
    async def GetTracksMetadata(self, track_ids: 'ao') -> 'aa{sv}':
        """Get all the metadata available for a set of tracks.

        Each set of metadata must have a "mpris:trackid" entry at the very
        least, which contains a string that uniquely identifies this track
        within the scope of the tracklist.

        Parameters
        ----------
        [track_ids]
            The list of track ids for which the metadata is requested.

        Returns
        -------
        [{str: MetaData}]
            List with the metadata of the tracks given as input.

        """
        logger.info(
            'Method '
            'org.mpris.MediaPlayer2.TrackList.GetTracksMetadata(track_ids=%s) '
            'called',
            track_ids)
        # Get list of mopidy ids
        track_ids = [self.dbusid_to_mopidyid(track_id)
                     for track_id in track_ids]
        # Get list of available tracks
        tl_tracks = await self.mopidy.tracklist.get_tl_tracks()
        # Convert list of available tracks to a dictionary
        tl_tracks = {tl_track['tlid']: tl_track
                     for tl_track in tl_tracks
                     if tl_track['tlid'] in track_ids}

        metadata = []
        for track_id in track_ids:
            md = await self.player_interface._get_Metadata(
                tl_track=tl_tracks[track_id])
            metadata.append(md)

        return metadata

    # AddTrack(s: Uri, o: AfterTrack, b: SetAsCurrent)  → nothing

    @method()
    async def AddTrack(self, uri: 's', after_track: 'o', set_as_current: 'b'):
        """Add a 'uri' in the 'TrackList'.

        Clients should not assume that the track has been added at the time
        when this method returns. They should wait for a 'TrackAdded' (or
        'TrackListReplaced') signal.

        If the 'CanEditTracks()' property is 'False', this has no effect.

        Parameters
        ----------
        uri : str
            The URI of the item to add. Its uri scheme should be an element
            of the "org.mpris.MediaPlayer2.SupportedUriSchemes" property and
            the mime-type should match one of the elements of the
            "org.mpris.MediaPlayer2.SupportedMimeTypes".
        after_track : TrackId
            The identifier of the track after which the new item should be
            inserted. The path "/org/mpris/MediaPlayer2/TrackList/NoTrack"
            indicates that the track should be inserted at the start of the
            track list.
        set_as_current : bool
            Whether the newly inserted track should be considered as the
            current track. Setting this to 'True' has the same effect as
            calling 'GoTo()' afterwards.

        Raises
        ------
        NotSupported
            May be raised, if the 'CanEditTracks' property is 'False'.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.TrackList.AddTrack'
            '(uri=%s, after_track=%s, set_as_current=%s) called',
            uri, after_track, set_as_current)
        if self._get_CanEditTracks():
            # TODO: Check if the uri is in 'SupportedUriSchemes' and the mime
            # type is in 'SupportedMimeTypes'
            if after_track == '/org/mpris/MediaPlayer2/TrackList/NoTrack':
                at = None
            else:
                # Mopidy id
                after_track = self.dbusid_to_mopidyid(after_track)
                at = await self.mopidy.tracklist.index(tlid=after_track) + 1
            await self.mopidy.tracklist.add(uris=[uri], at_position=at)
            await self.mopidy.playback.play(tlid=at)
            metadata = await self.player_interface._get_metadata()
            self.TrackAdded(metadata, after_track)
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot add a track, '
                f'because mopidy cannot edit the tracklist')

    # RemoveTrack(o: TrackId)                           → nothing

    @method()
    async def RemoveTrack(self, track_id: 'o'):
        """Remove an item from the 'TrackList'.

        Clients should not assume that the track has been removed at the
        time when this method returns. They should wait for a 'TrackRemoved'
        (or 'TrackListReplaced') signal.

        If the track is not part of this tracklist, this has no effect.

        If the 'CanEditTracks' property is 'False', this has no effect.

        Parameters
        ----------
        track_id : TrackId
            Identifier of the track to be removed.

            "/org/mpris/MediaPlayer2/TrackList/NoTrack" is not a valid value
            for this argument.

        Raises
        ------
        NotSupported
            May be raised, if the 'CanEditTracks' property is 'False'.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.TrackList.RemoveTrack'
            '(track_id=%s) called',
            track_id)
        if self._get_CanEditTracks():
            # Mopidy id
            tl_id = self.dbusid_to_mopidyid(track_id)
            await self.mopidy.tracklist.remove({'tlid': [tl_id]})
            self.TrackRemoved(track_id)
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot remove a track, '
                f'because mopidy cannot edit the tracklist')

    # GoTo(o: TrackId)                                  → nothing

    @method()
    async def GoTo(self, track_id: 'o'):
        """Skip to the specified 'track_id'.

        If the track is not part of this tracklist, this has no effect.

        If this object is not "/org/mpris/MediaPlayer2", the current
        'TrackList''s tracks should be replaced with the contents of this
        'TrackList', and the 'TrackListReplaced' signal should be fired from
        "/org/mpris/MediaPlayer2".

        Signals
        -------
        org.mpris.MediaPlayer2.TrackListReplaced
            Emit, if the current TrackList's tracks were replaced.

        Parameters
        ----------
        track_id : TrackId
            Identifier of the track to skip to.

            "/org/mpris/MediaPlayer2/TrackList/NoTrack" is not a valid value
            for this argument.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.TrackList.GoTo'
            '(track_id=%s) called',
            track_id)
        tracks = await self._get_Tracks()
        # Mopidy id
        tl_id = self.dbusid_to_mopidyid(track_id)
        if track_id in tracks:
            await self.mopidy.playback.play(tlid=tl_id)
            self.TrackListReplaced(tracks, track_id)
        else:
            pass

    #
    # Signals
    #

    # TrackListReplaced(ao: Tracks, o: CurrentTrack)

    @signal()
    def TrackListReplaced(self, tracks: 'ao', current_track: 'o') -> 'aoo':
        """Indicate that the entire tracklist has been replaced.

        It is left up to the implementation to decide when a change to the
        track list is invasive enough that this signal should be emitted
        instead of a series of 'TrackAdded' and 'TrackRemoved' signals.

        Parameters
        ----------
        tracks : [TrackId]
            The new content of the tracklist.
        current_track : TrackId
            The identifier of the track to be considered as current.

            "/org/mpris/MediaPlayer2/TrackList/NoTrack" indicates that there
            is no current track.

            This should correspond to the "mpris:trackid" field of the
            'Metadata' property of the "org.mpris.MediaPlayer2.Player"
            interface.

        Returns
        -------
        [[TrackId], TrackId]
            '[TrackId]':
                A list with the new content of the tracklist.
            'TrackId':
                The identifier of the track to be considered as current.

        """
        logger.info(
            'Signal org.mpris.MediaPlayer2.TrackList.TrackListReplaced'
            '(tracks=%s, current_track=%s) to emit',
            tracks, current_track)
        return [tracks, current_track]

    # TrackAdded(a{sv}: Metadata, o: AfterTrack)

    @signal()
    def TrackAdded(self, metadata: 'a{sv}', after_track: 'o') -> 'a{sv}o':
        """Indicate that a track has been added to the track list.

        Parameters
        ----------
        metadata : MetaData
            The metadata of the newly added item.

            This must include a "mpris:trackid" entry.
        after_track : TrackId
            The identifier of the track after which the new track was
            inserted. The path "/org/mpris/MediaPlayer2/TrackList/NoTrack"
            indicates that the track was inserted at the start of the track
            list.

        Returns
        -------
        [MetaData, TrackId]
            'MetaData':
                The metadata of the newly added item.
            'TrackId':
                The identifier of the track after which the new track was
                inserted.

        """
        logger.info(
            'Signal org.mpris.MediaPlayer2.TrackList.TrackAdded'
            '(metadata=%s, after_track=%s) to emit',
            metadata, after_track)
        return [metadata, after_track]

    # TrackRemoved(o: TrackId)

    @signal()
    def TrackRemoved(self, track_id: 'o') -> 'o':
        """Indicate that a track has been removed from the track list.

        Parameters
        ----------
        track_id : TrackId
            The identifier of the track being removed.

            "/org/mpris/MediaPlayer2/TrackList/NoTrack" is not a valid value
            for this argument.

        Returns
        -------
        TrackId
            The identifier of the track being removed.

        """
        logger.info(
            'Signal org.mpris.MediaPlayer2.TrackList.TrackRemoved'
            '(track_id=%s) to emit',
            track_id)
        return track_id

    # TrackMetadataChanged(o: TrackId, a{sv}: Metadata)

    @signal()
    def TrackMetadataChanged(self, track_id: 'o', metadata: 'a{sv}') -> 'oa{sv}':
        """Indicate that the metadata of a track in the tracklist has changed.

        This may indicate that a track has been replaced, in which case the
        "mpris:trackid" metadata entry is different from the 'track_id'
        argument.

        Parameters
        ----------
        track_id : TrackId
            The id of the track which metadata has changed.

            If the track id has changed, this will be the old value.

            "/org/mpris/MediaPlayer2/TrackList/NoTrack" is not a valid value
            for this argument.
        metadata : MetaData
            The new track metadata.

            This must include a "mpris:trackid" entry. If the track id has
            changed, this will be the new value.

        Returns
        -------
        [TrackId, MetaData]:
            'TrackId'
                The id of the track which metadata has changed.
            'MetaData'
                The new track metadata.

        """
        logger.info(
            'Signal org.mpris.MediaPlayer2.TrackList.TrackMetadataChanged'
            '(track_id=%s, metadata=%s) to emit',
            track_id, metadata)
        return [track_id, metadata]

    #
    # Properties
    #

    # Tracks                ao (Track_Id_List)      Read only

    @dbus_property(access=PropertyAccess.READ)
    async def Tracks(self) -> 'ao':
        """List with 'TrackId's of each track, in order.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted every time this property changes, but the
            signal message does not contain the new value. Client
            implementations should rather rely on the 'TrackAdded',
            'TrackRemoved' and 'TrackListReplaced' signals to keep their
            representation of the tracklist up to date.

        Returns
        -------
        [TrackId]
            A list with the track ids in the tracklist, in order.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.TrackList.Tracks requested')
        track_ids = await self._get_Tracks()
        return [self.mopidyid_to_dbusid(track_id) for track_id in track_ids]

    async def _get_Tracks(self):
        tl_tracks = await self.mopidy.tracklist.get_tl_tracks()
        return [tl_track['tlid'] for tl_track in tl_tracks]

    # CanEditTracks         b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    def CanEditTracks(self) -> 'b':
        """Whether 'AddTrack()' and 'RemoveTrack()' can be called.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            False:
                Calling 'AddTrack()' or 'RemoveTrack()' will have no effect
                (and may raise a 'NotSupported' error).
            True:
                'AddTrack()' and 'RemoveTrack()' can be called.

        """
        logger.info(
            'Property '
            'org.mpris.MediaPlayer2.TrackList.CanEditTracks requested')
        return self._get_CanEditTracks()

    def _get_CanEditTracks(self):
        return True



#!/usr/bin/env python3
"""Command line program for 'mopidris', a MPRIS D-Bus interface for Mopidy.

Usage
=====

Call with:
    python3 mopidris

See the help with all options explained:
    python3 mopidris --help

See also:
    - https://pypi.org/project/mopidris
    - https://codeberg.org/sph/mopidris

Please report bugs at:
    - https://codeberg.org/sph/mopidris/issues


Configuration File
------------------

`mopidris` reads a configuration file `mopidris.conf` (by default from the
default place for the user's configurations). It has a simple `key = value`
format and the keys are explained in the output of `python3 mopidris --help`.


Required Packages
=================

dbus-next >= 0.2.3
mopidy-asyncio-client >= 3.0.1


Optional Packages
=================

`mopidris` makes use of the following optional package, if it can find them,
but it also works well without them:
    - aiorun
    - colored_traceback
    - colorlog

For more information see https://pypi.org/project/mopidris or
https://codeberg.org/sph/mopidris


Created on Sun Mar 28 18:18:06 2021


Copyright (C) 2021  Stephan Helma

This file is part of mopidris.

mopidris is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mopidris is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mopidris. If not, see <https://www.gnu.org/licenses/>.

"""

# Import colored_traceback, so that we have it available for all other imports
try:
    # https://github.com/staticshock/colored-traceback.py
    # sudo pip3 install colored-traceback
    import colored_traceback
except ImportError:
    pass
else:
    colored_traceback.add_hook()


import os
from pathlib import Path
import sys
import argparse
import logging

# Import aiorun or asyncio
try:
    import aiorun
except ImportError:
    logging.warning(
        "Module 'aiorun' is not available. "
        "You might experience unpleasant sideeffects - such as locked ports - "
        "when shutting down.")
    import asyncio

from mopidris import main_loop


logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(name)s: %(message)s',
    datefmt='%H:%M:%S')


# Default values for command line options
__HOST = 'localhost'
__PORT = 6680
__ATTEMPTS = 5
__TIMEOUT = 20
__INTERFACES = []
__LOGLEVEL = logging.getLevelName(logging.getLogger().level).lower()


def get_userconfigfile():
    """Get the user's default configuration name.

    Returns
    -------
    str
        The name of the default user's configuration file.

    """
    # Get our name
    name = Path(sys.argv[0]).name
    # Get location of user's configuration directory
    if sys.platform == 'win32':
        # Borrowed from https://github.com/barry-scott/config-path/
        import ctypes
        import ctypes.wintypes

        buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(
            0, 0x1a,    # CSIDL_APPDATA, Application Data
            0, 0,       # SHGFP_TYPE_CURRENT, want current, not default value
            buf)

        config_folder = Path(buf.value)

    elif sys.platform == 'darwin':
        config_folder = Path('~/Library/Application Support/')
    else:
        config_folder = Path(os.getenv('XDG_DATA_HOME', '~/.config/'))

    return config_folder.joinpath(name, f'{name}.conf')


def parse_args(modules=[]):
    """Parse the command line arguments.

    Parameters
    ----------
    modules : [str], optional
        A list with the names of the modules, for which command line
        arguments should be automatically created in the 'debug' group, to
        set their logging levels independently of the main logging level.
        The default is [].

    Returns
    -------
    argparse.Namespace, {str: int}
        argparse.Namespace:
            The parsed command line arguments.
        {str: int}:
            Logging levels for modules. The key is the module name and the
            value the logging level.

    """
    # Help formatter
    class LinebreakDescriptionHelpFormatter(argparse.HelpFormatter):
        """Formatter which retains any line breaks in descriptions."""

        def _fill_text(self, text, width, indent):
            import textwrap
            return '\n'.join([
                textwrap.fill(
                    self._whitespace_matcher.sub(' ', paragraph).strip(),
                    width,
                    initial_indent=indent,
                    subsequent_indent=indent)
                for paragraph in text.splitlines()])

    # Type checkers
    def positiveint(string):
        """Parse 'string' into a positive integer."""
        i = int(string)
        if i > 0:
            return i
        else:
            raise ValueError('integer must be >0')

    def readablefile(string):
        """Check if 'string' is a readable file."""
        file = Path(string)
        if file.expanduser().is_file():
            return file
        else:
            raise ValueError(
                f"File '{string}' does not exist or is not readable")

    #
    # Command line parser
    #

    parser = argparse.ArgumentParser(
        description='Connect a (remote) Mopidy instance to the local DBus.',
        epilog='Configuration file:\n'
               "Each line in the configuration file is a 'key = value' pair. "
               "The key is given in the description of each parameter "
               "and the value can be any valid value.\n"
               "Comments start with '#' and everything after is discarded.\n"
               "\n"
               "Copyright (C) 2021, licensed under GPL 3.0 or later.",
        formatter_class=LinebreakDescriptionHelpFormatter)

    config_file = get_userconfigfile()

    #
    # Common arguments
    #

    parser.add_argument(
        nargs='?', default=__HOST,
        dest='host', metavar='HOST',
        help='the remote host name or ip address '
             '(default: %(default)s, configfile key: %(dest)s)')

    parser.add_argument(
        '-p', '--port', default=__PORT,
        dest='port', metavar='PORT',
        type=positiveint,
        help='the port number Mopidy is listening on '
             '(default: %(default)s, configfile key: %(dest)s)')

    parser.add_argument(
        '-r', '--retries', default=__ATTEMPTS,
        type=int,
        dest='reconnect_attempts', metavar='ATTEMPTS',
        help='how often to retry to connect to Mopidy; if -1, retry forever '
             '(default: %(default)s, configfile key: %(dest)s)')

    parser.add_argument(
        '-t', '--timeout', default=__TIMEOUT,
        type=positiveint,
        dest='reconnect_timeout', metavar='TIMEOUT',
        help='how long to wait between reconnection attempts '
             '(default: %(default)s, configfile key: %(dest)s)')

    parser.add_argument(
        '-q', '--can-quit',
        action='store_true',
        dest='can_quit',
        help='it is possible to quit this program from an application; '
             'note, that the (remote) will keep running '
             '(default: %(default)s, configfile key: %(dest)s)')

    parser.add_argument(
        '-m', '--no-mangle',
        action='store_false',
        dest='mangle',
        help='do not split the Internet radio stream title at " - " '
             '(default: %(default)s, configfile key: %(dest)s)')

    parser.add_argument(
        '-c', '--config',
        type=readablefile,
        dest='config', metavar='FILE',
        help=f'read values from the configuration file '
             f'(default: {config_file}, a configfile key is not supported)')

    #
    # DBus Interfaces
    #

    group = parser.add_argument_group('DBus Interfaces')

    group.add_argument(
        '-i', '--interface', default=__INTERFACES,
        choices=['tracklist', 'playlists'], action='append',
        dest='interfaces', metavar='{tracklist,playlists}',
        help="activate additional 'org.mpris.MediaPlayer2' DBus interfaces "
             "(the 'org.mpris.MediaPlayer2' and "
             "'org.mpris.MediaPlayer2.Player' interfaces are always active) "
             "(configfile key: %(dest)s - use one line for each interface)")

    #
    # Debugging
    #

    group = parser.add_argument_group('Debugging')

    # Available loglevels
    loglevels = [level[1].lower()
                 for level in sorted(logging._levelToName.items())
                 if level[0] > 0]

    group.add_argument(
        '-d', '--debug', default=__LOGLEVEL,
        choices=loglevels,
        dest='debug',
        help='the general loglevel '
             '(default: %(default)s, configfile key: %(dest)s)')

    # Do it for all given modules
    for module in modules:
        module_short = module.split('_', 1)[0]
        group.add_argument(
            f'--debug-{module_short}',
            choices=loglevels,
            dest=f'debug_{module}',
            help=f"the loglevel for the '{module}' module "
                 f"(default: follows --debug argument, "
                 f"configfile key: %(dest)s)")

    #
    # Parse and process
    #

    # 1st pass to get the configuration file
    args = parser.parse_args()

    if args.config is not None:
        config_file = args.config
    try:
        # Parse the configuration file and build a dictionary with its values,
        # discarding all values, which are not command line parameters
        default_args = {}
        with open(config_file.expanduser()) as f:
            for line in f:

                # Remove comments
                line = line.split('#', 1)[0].strip()
                if not line:
                    # Empty line
                    continue

                # Get keys and values
                try:
                    key, value = [t.strip() for t in line.split('=', 1)]
                except ValueError:
                    parser.error(
                        f"In the configuration file, "
                        f"the line '{line}' is missing the '=' sign")
                # No empty values are allowed
                if not value:
                    parser.error(
                        f"In the configuration file, "
                        f"the line '{line}' is missing a value")
                # Add it to our default args
                if key in args:
                    if key in default_args:
                        # Append value
                        v = default_args[key]
                        if isinstance(v, list):
                            # Append value to existing list
                            default_args[key].append(value)
                        else:
                            # Make a list
                            default_args[key] = [default_args[key], value]
                    else:
                        # Store value
                        default_args[key] = value
                else:
                    parser.error(
                        f"The key '{key}' found in the configuration file "
                        f"is not supported")
        # This is not very elegant, but there is no stable argparse API to
        # check the action of an argument (we could only use
        # 'parser._actions'), but we have to ensure that 'interfaces' is a
        # list!
        try:
            if not isinstance(default_args['interfaces'], list):
                default_args['interfaces'] = [default_args['interfaces']]
        except KeyError:
            pass

        # Set the parser's default arguments
        parser.set_defaults(**default_args)

    except FileNotFoundError:
        # Since argparse checked the existence for a user supplied
        # configuration file for us, this can only happens, if either
        #   - the default file should be used, but it does not exist, or
        #   - the user supplied file vanished between checking and opening
        # The first cause is ok, and we do not care about the second.
        pass

    except PermissionError as e:
        print(f'Could not open the configuration file.\n{e}', file=sys.stderr)
        exit(1)

    # 2nd pass, now that we have injected values from the configuration file
    args = parser.parse_args()

    # Remove the configuration file from the args namespace, because we
    # have dealt with it
    delattr(args, 'config')

    # mopidy_asyncio_client.MopidyClient requires 'None' and not negative
    # integer for infinite reconnection attempts
    if args.reconnect_attempts < 0:
        args.reconnect_attempts = None

    # Convert loglevel name for 'debug' to a numeic value
    args.debug = getattr(logging, args.debug.upper())

    # Transfer the modules' loglevels from the argparse.Namespace to a
    # separate dictionary (and convert it to number)
    module_loglevels = {}
    for module in modules:
        level = getattr(args, f'debug_{module}')
        if level is not None:
            # Store numeric loglevel for module, if given
            module_loglevels[module] = getattr(logging, level.upper())
        # Remove the attribute from the args namespace
        delattr(args, f'debug_{module}')

    return args, module_loglevels


if __name__ == '__main__':

    #
    # Process command line arguments
    #

    # Modules we want be able to change their logging level
    modules = ['mopidy_asyncio_client', 'dbus_next', 'websockets', 'aiorun',
               'interfaces.base', 'interfaces.root', 'interfaces.player',
               'interfaces.tracklist', 'interfaces.playlist']
    args, modules_loglevels = parse_args(modules=modules)

    #
    # Set loglevels (and minimum loglevel)
    #

    # Main logger
    loglevel_min = args.debug
    logging.getLogger().setLevel(args.debug)
    logging.info(
        'Loglevel set to: %s',
        logging.getLevelName(args.debug))

    # Module loggers
    for module, loglevel in modules_loglevels.items():
        loglevel_min = min(loglevel_min, loglevel)
        # Set loglevel for module, if given
        logging.getLogger(module).setLevel(loglevel)
        logging.info(
            "Loglevel for module '%s' set to: %s",
            module, logging.getLevelName(loglevel))

    #
    # Set debug
    #

    # We don't need logging level, only 'True' or 'False'
    args.debug = (args.debug == logging.DEBUG)

    #
    # Colorize log messages, if we any any loglevel equal or below debug level
    #
    if loglevel_min <= logging.DEBUG:
        try:
            import colorlog
        except ImportError:
            pass
        else:
            # Create color formatter
            # Would be so much easier, if this PR were accepted:
            # https://github.com/borntyping/python-colorlog/pull/99
            formatter = colorlog.ColoredFormatter(
                '%(asctime)s '
                '%(log_color)s%(levelname)-8s%(reset)s '
                '%(bold)s%(name)s%(reset)s: '
                '%(message_log_color)s%(message)s%(reset)s',
                datefmt='%H:%M:%S',
                log_colors={
                    'DEBUG':    'cyan',
                    'INFO':     'green',
                    'WARNING':  'yellow',
                    'ERROR':    'red',
                    'CRITICAL': 'bold_red'},
                secondary_log_colors={
                    'message': {
                        'ERROR':    'red',
                        'CRITICAL': 'bold_red'}})
            # Replace formatters of all stream handlers
            for handler in logging.getLogger().handlers:
                if isinstance(handler, logging.StreamHandler):
                    handler.setFormatter(formatter)

    #
    # Start asyncio loop
    #

    # Log start up info
    logging.info(
        f'Connect to mopidy on {args.host}:{args.port}')
    _attempts = ('forever'
                 if args.reconnect_attempts is None
                 else f'{args.reconnect_attempts} times')
    logging.info(
        f'If connection fails or breaks, '
        f'try to (re)connect {_attempts} '
        f'with {args.reconnect_timeout}s between retries')
    logging.info(
        f'Load additional DBus interfaces: '
        f'{", ".join(args.interfaces) or None}')

    # Convert args to a dictionary, so that we can use **kwargs
    kwargs = vars(args)

    while True:
        try:
            # Start either aiorun or asyncio event loop
            try:
                aiorun
            except NameError:
                # Run with standard asyncio event loop
                try:
                    loop = asyncio.get_event_loop()
                    loop.run_until_complete(
                        main_loop(loop_type='asyncio', **kwargs))
                except asyncio.CancelledError:
                    pass
            else:
                # Run forever with aiorun (stop only in debugging mode)
                aiorun.run(
                    main_loop(loop_type='aiorun', **kwargs),
                    stop_on_unhandled_errors=(args.debug))
            break
        except OSError:
            # Network connection error?
            import time
            time.sleep(30)



"""The Player MPRIS D-Bus interface for a remote Mopidy instance.

The D-Bus name of this interface is 'org.mpris.MediaPlayer2.Player'.


Created on Sun Mar 28 18:18:06 2021


Copyright (C) 2021  Stephan Helma

This file is part of mopidris.

mopidris is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

mopidris is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with mopidris. If not, see <https://www.gnu.org/licenses/>.

"""

import logging

import dbus_next as dbus
from dbus_next.service import method, dbus_property, PropertyAccess, signal

from .base import _MopidyBaseInterface


logger = logging.getLogger(__name__)


class MopidyPlayerInterface(_MopidyBaseInterface):
    """The "org.mpris.MediaPlayer2.Player" interface.

    This interface implements the methods for querying and providing basic
    control over what is currently playing.

    See:
        https://specifications.freedesktop.org/mpris-spec/2.2/Player_Interface.html#Property:Rate
    """

    def __init__(self, mopidy, mangle=True, *args, **kwargs):
        """Initialize the 'MopidyPlayerInterface' class.

        Parameters
        ----------
        mopdiy : mopidy_asyncio_client.MopidyClient
            The mopidy client
        mangle : bool
            If the titles of Internet radio streams should be split at " - "
            into artist and title.
        *args : arguments
            Passed on to parent class.
        **kwargs : keyword arguments
            Passed on to parent class.

        Returns
        -------
        None.

        """
        self._mangle = mangle
        super().__init__('org.mpris.MediaPlayer2.Player',
                         mopidy, *args, **kwargs)

    #
    # Mopidy events
    #

    async def on_mopidyevent_mute_changed(self, data):
        """Is called whenever the mute state is changed.

        Parameters
        ----------
        data['mute'] : boolean
            The new mute state.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.Player.mute_changed'
            '(data=%s) received',
            data)
        if data['mute']:
            volume = await self._get_Volume()
        else:
            volume = 0
        self.emit_properties_changed(
            Volume=volume)

    async def on_mopidyevent_options_changed(self, data):
        """Is called whenever an options is changed.

        Returns
        -------
        None.

        """
        loop = await self._get_LoopStatus()
        shuffle = await self._get_Shuffle()
        can_go_previous = await self._get_CanGoPrevious()
        can_go_next = await self._get_CanGoNext()
        self.emit_properties_changed(
            LoopStatus=loop,
            Shuffle=shuffle,
            CanGoPrevious=can_go_previous,
            CanGoNext=can_go_next)

    async def on_mopidyevent_playback_state_changed(self, data):
        """Is called whenever playback state is changed.

        Parameters
        ----------
        data['old_state'] : string from mopidy.core.PlaybackState field
            The state before the chan.
        data['new_state'] : string from mopidy.core.PlaybackState field
            The state after the change.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.Player.playback_state_changed'
            '(data=%s) received',
            data)
        status = data['new_state']
        self.emit_properties_changed(
            PlaybackStatus=status.capitalize())
        if status == 'playing':
            metadata = await self._get_Metadata()
            self.emit_properties_changed(
                PlaybackStatus=status.capitalize(),
                Metadata=metadata)

    async def on_mopidyevent_stream_title_changed(self, data):
        """Is called whenever the currently playing stream title changes.

        Parameters
        ----------
        data['title'] : string
            The new stream title.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.Player.stream_title_changed'
            '(data=%s) received',
            data)
        metadata = await self._get_Metadata()
        self.emit_properties_changed(
            Metadata=metadata)

    async def on_mopidyevent_seeked(self, data):
        """Is called whenever the time position changes by an unexpected amount.

        E.g. at seek to a new time position.

        Parameters
        ----------
        data['time_position'] : int
            The position that was seeked to in milliseconds.

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.Player.seeked'
            '(data=%s) received',
            data)
        # in µs
        self.Seeked(data['time_position'] * 1000)

    async def on_mopidyevent_volume_changed(self, data):
        """Is called whenever the volume is changed.

        Parameters
        ----------
        data ['volume'] : int
             The new volume in the range [0..100].

        Returns
        -------
        None.

        """
        logger.info(
            'Mopidy event org.mpris.MediaPlayer2.Player.volume_changed'
            '(data=%s) received',
            data)
        self.emit_properties_changed(
            Volume=data['volume']/100)

    #
    # Methods
    #

    # Next()                                            → nothing

    @method()
    async def Next(self):
        """Skip to the next track in the tracklist.

        If there is no next track (and endless playback and track repeat are
        both off), stop playback.

        If playback is paused or stopped, it remains that way.

        If 'CanGoNext' is 'False', attempting to call this method should have
        no effect.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.Next() called')
        if await self._get_CanGoNext():
            await self.mopidy.playback.next()
        else:
            pass

    # Previous()                                        → nothing

    @method()
    async def Previous(self):
        """Skip to the previous track in the tracklist.

        If there is no previous track (and endless playback and track repeat
        are both off), stop playback.

        If playback is paused or stopped, it remains that way.

        If 'CanGoPrevious' is 'False', attempting to call this method should
        have no effect.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.Previous() called')
        if await self._get_CanGoPrevious():
            await self.mopidy.playback.previous()
        else:
            pass

    # Pause()                                           → nothing

    @method()
    async def Pause(self):
        """Pause playback.

        If playback is already paused, this has no effect.

        Calling 'Play()' after this should cause playback to start again from
        the same position.

        If 'CanPause' is 'False', attempting to call this method should have
        no effect

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.Pause() called')
        if self._get_CanPause():
            await self.mopidy.playback.pause()
        else:
            pass

    # PlayPause()                                       → nothing

    @method()
    async def PlayPause(self):
        """Pause playback.

        If playback is already paused, resumes playback.

        If playback is stopped, starts playback.

        If 'CanPause' is 'False', attempting to call this method should have
        no effect and raise an error.

        Raises
        ------
        NotSupported
            Raised, if the 'CanPause' property is 'False'.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.PlayPause() called')
        if self.CanPause:
            state = await self._get_PlaybackStatus()
            if state == 'playing':
                await self.mopidy.playback.pause()
            elif state == 'paused':
                if await self._get_CanPlay():
                    await self.mopidy.playback.resume()
            elif state == 'stopped':
                await self.mopidy.playback.play()
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot pause')

    # Stop()                                            → nothing

    @method()
    async def Stop(self):
        """Stop playback.

        If playback is already stopped, this has no effect.

        Calling 'Play()' after this should cause playback to start again from
        the beginning of the track.

        If 'CanControl' is 'False', attempting to call this method should have
        no effect and raise an error.

        Raises
        ------
        NotSupported
            Raised, if the 'CanControl' property is 'False'.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.Stop() called')
        if self._get_CanControl():
            await self.mopidy.playback.stop()
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot stop')

    # Play()                                            → nothing

    @method()
    async def Play(self):
        """Start or resumes playback.

        If already playing, this has no effect.

        If paused, playback resumes from the current position.

        If there is no track to play, this has no effect.

        If 'CanPlay' is 'False', attempting to call this method should have
        no effect.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.Play() called')
        if await self._get_CanPlay():
            state = await self._get_PlaybackStatus()
            if state == 'paused':
                await self.mopidy.playback.resume()
            else:
                await self.mopidy.playback.play()
        else:
            pass

    # Seek(x: Offset in µs)                             → nothing

    @method()
    async def Seek(self, offset: 'x'):
        """Seek forward in the current track.

        If the 'CanSeek' property is 'False', this has no effect.

        Parameters
        ----------
        offset : float
            The number of microseconds to seek forward.

            A negative value seeks back. If this would mean seeking back
            further than the start of the track, the position is set to 0.

            If the value passed in would mean seeking beyond the end of the
            track, acts like a call to 'Next()'.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.Seek'
            '(offset=%s) called',
            offset)
        if await self._get_CanSeek():
            # in ms
            position = await self._get_Position()
            position += offset // 1000
            position = min(0, position)
            # If position would mean to seek beyond the end of the track,
            # act like a call to Next()
            tl_track = await self.mopidy.playback.get_current_tl_track()
            if position <= tl_track['track']['length']:
                await self.mopidy.playback.seek(position)
            else:
                await self.mopidy.playback.next()
        else:
            pass

    # SetPosition(o: TrackId, x: Position in µs)        → nothing

    @method()
    async def SetPosition(self, track_id: 'o', position: 'x'):
        """Set the current track position.

        If the 'position' argument is less than 0, do nothing.

        If the 'position' argument is greater than the track length,
        do nothing.

        If the 'CanSeek' property is 'False', this has no effect.

        Parameters
        ----------
        track_id : TrackId
            The currently playing track's identifier.

            If this does not match the id of the currently-playing track,
            the call is ignored as "stale".
        position : float (0 ≤ position ≤ <track_length>)
            Track position in microseconds.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.SetPosition'
            '(track_id=%s, position=%s) called',
            track_id, position)
        if self.CanSeek:

            if position < 0:
                return

            # in ms
            position //= 1000
            tl_track = await self.mopidy.playback.get_current_tl_track()

            if tl_track is None:
                return
            if track_id != self.tltrack_to_dbusid(tl_track):
                return
            if tl_track.track.length < position:
                return
            await self.mopidy.playback.seek(position)
        else:
            pass

    # OpenUri(s: Uri)                                   → nothing

    @method()
    async def OpenUri(self, uri: 's'):
        """Open the 'uri'.

        If the playback is stopped, starts playing.

        If the uri scheme or the mime-type of the uri to open is not supported,
        this method does nothing and may raise an error. In particular, if the
        list of available uri schemes is empty, this method may not be
        implemented.

        Clients should not assume that the 'uri' has been opened as soon as
        this method returns. They should wait until the "mpris:trackid" field
        in the 'Metadata' property changes.

        If the media player implements the 'TrackList' interface, then the
        opened track should be made part of the tracklist, the
        "org.mpris.MediaPlayer2.TrackList.TrackAdded" or
        "org.mpris.MediaPlayer2.TrackList.TrackListReplaced" signal should be
        fired, as well as the
        "org.freedesktop.DBus.Properties.PropertiesChanged" signal on the
        tracklist interface.

        Parameters
        ----------
        uri : str
            'uri' of the track to load. Its uri scheme should be an element
            of the "org.mpris.MediaPlayer2.SupportedUriSchemes" property and
            the mime-type should match one of the elements of the
            "org.mpris.MediaPlayer2.SupportedMimeTypes".

        Raises
        ------
        NotSupported
            May be raised, if the uri scheme or the mime-type of the uri to
            open is not supported.

        Returns
        -------
        None.

        """
        logger.info(
            'Method org.mpris.MediaPlayer2.Player.OpenUri'
            '(uri=uri) called',
            uri)
        tl_tracks = await self.mopidy.tracklist.add(uris=[uri])
        if tl_tracks:
            await self.mopidy.playback.play(tlid=tl_tracks[0].tlid)
            # dbus.DBusError(
            #     dbus.ErrorType.NOT_SUPPORTED,
            #     f'{self.mopidy} does not support uri {uri}')
            # dbus.DBusError(
            #     dbus.ErrorType.NOT_SUPPORTED,
            #     f'{self.mopidy} cannot mime type {XXX}')

    #
    # Signals
    #

    # Seeked(x: Position)

    @signal()
    def Seeked(self, val: 'x') -> 'x':
        """Indicate that the track position has changed in an inconsistant way.

        Indicates that the track position has changed in a way that is
        inconsistant with the current playing state.

        When this signal is not received, clients should assume that:
            - When playing, the position progresses according to the rate
              property.
            - When paused, it remains constant.

        This signal does not need to be emitted when playback starts or when
        the track changes, unless the track is starting at an unexpected
        position. An expected position would be the last known one when going
        from 'paused' to 'playing', and 0 when going from 'stopped' to
        'playing'.

        Parameters
        ----------
        val : int
            The new position, in microseconds.

        Returns
        -------
        int
            The new position in microseconds.

        """
        logger.info(
            'Signal org.mpris.MediaPlayer2.Seeked() to emit')
        return val

    #
    # Properties
    #

    # PlaybackStatus        s (Playback_Status)     Read only

    @dbus_property(access=PropertyAccess.READ)
    async def PlaybackStatus(self) -> 's':                      # noqa: D401
        """The current playback status.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        str
            "Playing":
                The media player is playing.
            "Paused":
                The media player is paused.
            "Stopped":
                The media player is stopped.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.PlaybackStatus requested')
        status = await self._get_PlaybackStatus()
        return status.capitalize()

    async def _get_PlaybackStatus(self):
        return await self.mopidy.playback.get_state()

    # LoopStatus            s (Loop_Status)         Read/Write (optional)

    @dbus_property(access=PropertyAccess.READWRITE)
    async def LoopStatus(self) -> 's':                          # noqa: D401
        """The current loop / repeat status.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Parameters
        ----------
        value : str
            "None":
                The playback will stop when there are no more tracks to play.
            "Track":
                The current track will start again from the begining once it
                has finished playing.
            "Playlist":
                The playback will loop through a list of tracks.

        Returns
        -------
        str
            "None":
                The playback stops when there are no more tracks to play.
            "Track":
                The current track starts again from the begining once it has
                finished playing.
            "Playlist":
                The playback loops through a list of tracks.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.LoopStatus requested')
        status = await self._get_LoopStatus()
        return str(status).capitalize()

    @LoopStatus.setter
    async def LoopStatus(self, value: 's'):
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.LoopStatus set to: %s',
            value)
        await self._set_LoopStatus(value)

    async def _get_LoopStatus(self):
        repeat = await self.mopidy.tracklist.get_repeat()
        if not repeat:
            return None

        single = await self.mopidy.tracklist.get_single()
        if single:
            return 'track'
        else:
            return 'playlist'

    async def _set_LoopStatus(self, value):
        if self._get_CanControl():
            if value is None:
                await self.mopidy.tracklist.set_repeat(False)
                await self.mopidy.tracklist.set_single(False)
            elif value == 'track':
                await self.mopidy.tracklist.set_repeat(True)
                await self.mopidy.tracklist.set_single(True)
            elif value == 'playlist':
                await self.mopidy.tracklist.set_repeat(True)
                await self.mopidy.tracklist.set_single(False)
            self.emit_properties_changed(
                LoopStatus=str(value).capitalize())
        else:
            pass

    # Rate                  d (Playback_Rate)       Read/Write

    @dbus_property(access=PropertyAccess.READWRITE)
    def Rate(self) -> 'd':                                      # noqa: D401
        """The current playback rate.

        If the media player has no ability to play at speeds other than the
        normal playback rate, this must still be implemented, and must return
        1.0. The 'MinimumRate' and 'MaximumRate' properties must also be set
        to 1.0.

        Not all values may be accepted by the media player. It is left to the
        media player implementations to decide how to deal with values they
        cannot use; they may either ignore them or pick a "best fit" value.
        Clients are recommended to only use sensible fractions or multiples
        of 1 (e.g.: 0.5, 0.25, 1.5, 2.0, etc).

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Parameters
        ----------
        value : float ('MinimumRate' < value < 'MaximumRate', ≠ 0)
            If playback is paused, the 'PlaybackStatus' property should be
            used to indicate this. A value of 0.0 should not be set by the
            client. If it is, the media player should act as though 'Pause()'
            was called.

        Returns
        -------
        float
            The current playback rate.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.Rate requested')
        return self._get_Rate()

    @Rate.setter
    async def Rate(self, value: 'd'):
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.Rate set to: %s',
            value)
        await self._set_Rate(value)

    def _get_Rate(self):
        return 1.

    async def _set_Rate(self, value):
        if self._get_CanPause() and value == 0:
            await self.mopidy.playback.pause()

    # Shuffle               b                       Read/Write (optional)

    @dbus_property(access=PropertyAccess.READWRITE)
    async def Shuffle(self) -> 'b':                             # noqa: D401
        """Whether the playback is progressing linearly through a playlist.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Parameters
        ----------
        value : bool
            'False':
                The playback should progress linearly through a playlist.
            'True':
                The playback should progress in some other order.

            If 'CanControl' is 'False', attempting to set this property
            should have no effect and raise an error.

        Raises
        ------
        NotSupported
            Raised, if the 'CanControl' property is 'False'.

        Returns
        -------
        bool
            'False':
                The playback is progressing linearly through a playlist.
            'True':
                The playback is progressing through a playlist in some other
                order.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.Shuffle requested')
        if self._get_CanControl():
            return await self._get_Shuffle()
        else:
            return False

    @Shuffle.setter
    async def Shuffle(self, value: 'b'):
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.Shuffle set to: %s',
            value)
        await self._set_Shuffle(value)

    async def _get_Shuffle(self):
        return await self.mopidy.tracklist.get_random()

    async def _set_Shuffle(self, value):
        if self._get_CanControl():
            value = bool(value)
            await self.mopidy.tracklist.set_random(value)
            self.emit_properties_changed(
                Shuffle=value)
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot shuffle, '
                f'because mopidy cannot be controlled')

    # Metadata              a{sv} (Metadata_Map)    Read only

    @dbus_property(access=PropertyAccess.READ)
    async def Metadata(self) -> 'a{sv}':                        # noqa: D401
        """The metadata of the current element.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        {str: MetaData}
            The dictionary with the metadata or an empty dictionary.

            If there is a current track, this must have a "mpris:trackid"
            entry (of D-Bus type "o") at the very least, which contains a
            D-Bus path that uniquely identifies this track.

            See the type documentation for common fields:
                https://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata/

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.Metadata requested')
        return await self._get_Metadata()

    async def _get_Metadata(self, tl_track=None):
        data = {}

        if tl_track is None:
            tl_track = await self.mopidy.playback.get_current_tl_track()
            if tl_track is None:
                return data

        track = tl_track['track']

        # Fill metadata dictionary

        full_track_id = self.tltrack_to_dbusid(tl_track)
        data['mpris:trackid'] = dbus.Variant('o', full_track_id)

        if length := track.get('length', None):
            data['mpris:length'] = dbus.Variant('x', length * 1000)

        if uri := track.get('uri', None):
            data['xesam:url'] = dbus.Variant('s', uri)

            images = await self.mopidy.library.get_images([uri])
            if images and images[uri]:
                largest_image = max(
                    images[uri],
                    key=lambda i: (i.width or 0, i.height or 0))
                data['mpris:artUrl'] = dbus.Variant('s', largest_image['uri'])

        stream_title = await self.mopidy.playback.get_stream_title()
        if title := stream_title or track.get('name', None):
            data['xesam:title'] = dbus.Variant('s', title)

        if artists := track.get('artists', None):
            # artists = list(track.artists)
            # artists.sort(key=lambda a: a.name or '')
            # data['xesam:artist'] = dbus.Variant(
            #     'as', [a.name for a in artists if a.name]
            # )
            data['xesam:artist'] = dbus.Variant(
                'as', [a['name'] for a in artists if 'name' in a])

        # Fix for radio streams, which only hava a title and no artist
        if title and artists is None:
            if self._mangle and (
                    uri.startswith('http')
                    or uri.startswith('orfradio:')):
                # Internet station?
                try:
                    artists, title = title.split(' - ', 1)
                except ValueError:
                    artists, title = None, None
            else:
                artists, title = None, None

            if title is not None and artists is not None:
                data['xesam:title'] = dbus.Variant('s', title)
                # Split on '&' and ','
                data['xesam:artist'] = dbus.Variant(
                    'as', [a.strip()
                           for a in artists.replace('&', ',').split(',')])

        if album := track.get('album', None):
            if name := album.get('name', None):
                data['xesam:album'] = dbus.Variant('s', name)

            if artists := album.get('artists', None):
                # artists = list(track.album.artists)
                # artists.sort(key=lambda a: a.name or '')
                # data['xesam:albumArtist'] = dbus.Variant(
                #     'as', [a.name for a in artists if a.name]
                # )
                data['xesam:albumArtist'] = dbus.Variant(
                    'as',
                    [artist['name'] for artist in artists if 'name' in artist])

        if disc_no := track.get('disc_no', None):
            data['xesam:discNumber'] = dbus.Variant('i', disc_no)

        if track_no := track.get('track_no', None):
            data['xesam:trackNumber'] = dbus.Variant('i', track_no)

        return data

    # Volume                d (Volume)              Read/Write

    @dbus_property(access=PropertyAccess.READWRITE)
    async def Volume(self) -> 'd':                              # noqa: D401
        """The volume level.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Parameters
        ----------
        value : float (0 ≤ value ≤ 1)
            The volume level. If a negative value is passed, the volume
            should be set to 0.0.

            If 'CanControl' is 'False', attempting to set this property
            should have no effect and raise an error.

        Raises
        ------
        NotSupported
            Raised, if the 'CanControl' property is 'False'.


        Returns
        -------
        float (0 ≤ value ≤ 1)
            The volume level.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.Volume requested')
        if self._get_CanControl():
            return await self._get_Volume()
        else:
            return 0.

    @Volume.setter
    async def Volume(self, value: 'd'):
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.Volume set to: %s',
            value)
        await self._set_Volume(value)

    async def _get_Volume(self):
        mute = await self.mopidy.mixer.get_mute()
        if mute:
            return 0.

        volume = await self.mopidy.mixer.get_volume()
        if volume is None:
            return 0.
        else:
            # Volume between 0 and 1
            return volume / 100.

    async def _set_Volume(self, value):
        if self._get_CanControl():
            # Volume between 0 and 100
            volume = int(max(0, min(1, value)) * 100)
            await self.mopidy.mixer.set_volume(int(volume*100))
            self.emit_properties_changed(
                Volume=volume/100)
            if volume:
                await self.mopidy.mixer.set_mute(False)
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot set volume, '
                f'because mopidy cannot be controlled')

    # Position              x (Time_In_Us)          Read only

    @dbus_property(access=PropertyAccess.READ)
    async def Position(self) -> 'x':                            # noqa: D401
        """The current track position in microseconds.

        If the media player allows it, the current playback position can be
        changed either with the 'SetPosition()' method or the 'Seek()' method
        on this interface. If this is not the case, the 'CanSeek' property is
        'False', and setting this property has no effect and can raise an
        error.

        Changes
        -------
            Do NOT emit the "org.freedesktop.DBus.Properties.PropertiesChanged"
            signal!

        Signals
        -------
        Seeked
            If the playback progresses in a way that is inconstistant with
            the 'Rate' property, the 'Seeked' signal is emitted.

        Raises
        ------
        NotSupported
            May be raised, if the 'CanSeek' property is 'False'.

        Returns
        -------
        int (0 ≤ value ≤ metadata['mpris:length'])
            The current track position in microseconds, between 0 and the
            "mpris:length" metadata entry (see 'Metadata').

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.Position requested')
        if await self._get_CanSeek():
            position = await self._get_Position()       # in ms
            return position * 1000                      # in µs
        else:
            dbus.DBusError(
                dbus.ErrorType.NOT_SUPPORTED,
                f'{self.mopidy} cannot get position, '
                f'because mopidy cannot be seeked')
            return 0

    async def _get_Position(self):
        # in ms
        return await self.mopidy.playback.get_time_position()

    # MinimumRate           d (Playback_Rate)       Read only

    @dbus_property(access=PropertyAccess.READ)
    def MinimumRate(self) -> 'd':                               # noqa: D401
        """The minimum value, which the 'Rate' property can take.

        Clients should not attempt to set the 'Rate' property below this value.

        Note that even if this value is 0.0 or negative, clients should not
        attempt to set the 'Rate' property to 0.0.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        float (value ≤ 1)
            The minimum value, which the 'Rate' property can take.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.MinimumRate requested')
        return self._get_MinimumRate()

    def _get_MinimumRate(self):
        return 1.

    # MaximumRate           d (Playback_Rate)       Read only

    @dbus_property(access=PropertyAccess.READ)
    def MaximumRate(self) -> 'd':                               # noqa: D401
        """The maximum value, which the 'Rate' property can take.

        Clients should not attempt to set the 'Rate' property above this value.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        float (value ≥ 1)
            The maximum value, which the 'Rate' property can take.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.MaximumRate requested')
        return self._get_MaximumRate()

    def _get_MaximumRate(self):
        return 1.

    # CanGoNext             b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    async def CanGoNext(self) -> 'b':
        """Whether the client can call the 'Next()' method.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                The client cannot call the 'Next()' method on this interface.
            'True':
                The client can call the 'Next()' method on this interface and
                expect the current track to change.

            If it is unknown, whether a call to 'Next()' will be successful
            (for example, when streaming tracks), this property should be set
            to 'True'.

            If the 'CanControl' property is 'False', this property should
            also be 'False'.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.CanGoNext requested')
        return await self._get_CanGoNext()

    async def _get_CanGoNext(self):
        if self._get_CanControl():
            current_tlid = await self.mopidy.playback.get_current_tlid()
            next_tlid = await self.mopidy.tracklist.get_next_tlid()
            return current_tlid != next_tlid
        else:
            return False

    # CanGoPrevious         b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    async def CanGoPrevious(self) -> 'b':
        """Whether the client can call the 'Previous()' method.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                The client cannot call the 'Previous()' method on this
                interface.
            'True':
                The client can call the 'Previous()' method on this
                interface and expect the current track to change.

            If it is unknown, whether a call to 'Previous()' will be
            successful (for example, when streaming tracks), this property
            should be set to 'True'.

            If the 'CanControl' property is 'False', this property should
            also be 'False'.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.CanGoPrevious requested')
        return await self._get_CanGoPrevious()

    async def _get_CanGoPrevious(self):
        if self._get_CanControl():
            current_tlid = await self.mopidy.playback.get_current_tlid()
            prev_tlid = await self.mopidy.tracklist.get_previous_tlid()
            return current_tlid != prev_tlid
        else:
            return False

    # CanPlay               b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    async def CanPlay(self) -> 'b':
        """Whether playback can be started using 'Play()' or 'PlayPause()'.

        Note that this is related to whether there is a "current track": the
        value should not depend on whether the track is currently paused or
        playing. In fact, if a track is currently playing (and 'CanControl' is
        'True'), this should be 'True'.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                The playback can not be started.
            'True':
                The playback can be started using 'Play()' or 'PlayPause()'.

            If 'CanControl' is 'False', this property should also be 'False'.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.CanPlay requested')
        return await self._get_CanPlay()

    async def _get_CanPlay(self):
        if self.CanControl:
            current_tlid = await self.mopidy.playback.get_current_tlid()
            if current_tlid is not None:
                return True
            next_tlid = await self.mopidy.tracklist.get_next_tlid()
            if next_tlid is not None:
                return True
            return False
        else:
            return False

    # CanPause              b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    def CanPause(self) -> 'b':
        """Whether playback can be paused using 'Pause()' or 'PlayPause()'.

        Note that this is an intrinsic property of the current track: its
        value should not depend on whether the track is currently paused or
        playing. In fact, if playback is currently paused (and 'CanControl' is
        'True'), this should be 'True'.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                The playback cannot be paused.
            'True':
                The playback can be paused using 'Pause()' or 'PlayPause()'.

            If 'CanControl' is 'False', this property should also be 'False'.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.CanPause requested')
        return self._get_CanPause()

    def _get_CanPause(self):
        if self._get_CanControl():
            return True
        else:
            return False

    # CanSeek               b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    async def CanSeek(self) -> 'b':
        """Whether the client can control the playback position using 'Seek()'
        and 'SetPosition()'.

        This may be different for different tracks.

        Signals
        -------
        org.freedesktop.DBus.Properties.PropertiesChanged
            This signal is emitted with the new value every time this
            property changes.

        Returns
        -------
        bool
            'False':
                The playback position cannot be set.
            'True':
                The playback position can be set using 'Seek()' and
                'SetPosition()'.

            If 'CanControl' is 'False', this property should also be 'False'.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.CanSeek requested')
        return await self._get_CanSeek()

    async def _get_CanSeek(self):
        if self._get_CanControl():
            # The track information has no information, if the media can be
            # seeked. A workaround is to use the '.length' information
            # (because) non-seekable streams usually do not have a length
            tl_track = await self.mopidy.playback.get_current_tl_track()
            if (tl_track
                and tl_track['track'].get('length', None) is not None
            ):
                return True
            else:
                return False
        else:
            return False

    # CanControl            b                       Read only

    @dbus_property(access=PropertyAccess.READ)
    def CanControl(self) -> 'b':
        """Whether the media player may be controlled over this interface.

        This property is not expected to change, as it describes an intrinsic
        capability of the implementation.

        Changes
        -------
            Do NOT emit the "org.freedesktop.DBus.Properties.PropertiesChanged"
            signal!

        Returns
        -------
        bool
            'False':
                Clients should assume that all properties on this interface
                are read-only (and will raise errors if writing to them is
                attempted), no methods are implemented and all other
                properties starting with "Can" are also 'False'.
            'True':
                The media player may be controlled over this interface.

        """
        logger.info(
            'Property org.mpris.MediaPlayer2.Player.CanControl requested')
        return self._get_CanControl()

    def _get_CanControl(self):
        return True
